#include <Arduino.h>
#include <stdint.h>
#include <ComponentObject.h>
#include <RangeSensor.h>
#include <SparkFun_VL53L1X.h>
#include <vl53l1x_class.h>
#include <vl53l1_error_codes.h>
#include "SCMD.h"
#include "SCMD_config.h" //Contains #defines for common SCMD register names and values
#include "Wire.h"
#include "SparkFun_VL53L1X.h"
#define LEDPIN 13
SCMD myMotorDriver; //This creates the main object of one motor driver and connected slaves.
SFEVL53L1X distanceSensor;
#include <math.h>
#include "ICM_20948.h"  // Click here to get the library: http://librarymanager/All#SparkFun_ICM_20948_IMU
//#define USE_SPI       // Uncomment this to use SPI
#define SERIAL_PORT Serial
#define SPI_PORT SPI    // Your desired SPI port.       Used only when "USE_SPI" is defined
#define CS_PIN 2        // Which pin you connect CS to. Used only when "USE_SPI" is defined
#define WIRE_PORT Wire  // Your desired Wire port.      Used when "USE_SPI" is not defined
#define AD0_VAL   1     // The value of the last bit of the I2C address. 
// On the SparkFun 9DoF IMU breakout the default is 1, and when
// the ADR jumper is closed the value becomes 0
#ifdef USE_SPI
ICM_20948_SPI myICM;  // If using SPI create an ICM_20948_SPI object
#else
ICM_20948_I2C myICM;  // Otherwise create an ICM_20948_I2C object
#endif

// maximum length of reply / data message
#define MAXREPLY 100
#define TODO_VAL 0
// buffer to reply to client
uint8_t val[MAXREPLY];
uint8_t *val_data = &val[2]; // start of optional data area
uint8_t *val_len = &val[1];  // store length of optional data
#include "BLE_example.h"
#include "commands.h"
#include "related_funcs.h"
String s_Rev = "Rev 1.0";
String s_Rcvd = "100"; //KHE declare extern in BLE_example_funcs.cpp to accept messages, if 100 don't bother checking case statements
uint16_t l_Rcvd = 0;
uint8_t *m_Rcvd = NULL;
String s_AdvName = "MyRobot"; //KHE 2 0 TOTAL CHARACHTERS ONLY!!  any more will be dropped

cmd_t empty_cmd = {NOT_A_COMMAND, 1, {0}};
cmd_t *cmd = &empty_cmd;
cmd_t *res_cmd = &empty_cmd;
bt_debug_msg_t *bt_debug_head = NULL;
bt_debug_msg_t *bt_debug_tail = NULL;
//SCMD motorDriver;
present_t presentSensors = {
  .motorDriver = 0,
  .ToF_sensor = 0,
  .prox_sensor = 0,
  .IMU = 0
};
int bytestream_active = 0;


//--------PID imports----------
#include <PID_v1.h>
double Setpoint, Input, Output;
double Kp = 20, Ki = 0, Kd = 0;
PID myPID(&Input, &Output, &Setpoint, Kp, Ki, Kd, DIRECT);

void setup()
{
  Serial.begin(115200);
  delay(1000);
  pinMode(LEDPIN, OUTPUT);
  myMotorDriver.settings.commInterface = I2C_MODE;
  myMotorDriver.settings.I2CAddress = 0x5D; //config pattern "0101" on board for address 0x5A
  myMotorDriver.settings.chipSelectPin = 10;
  while ( myMotorDriver.begin() != 0xA9 ) //Wait until a valid ID word is returned
  {
    delay(500);
  }
  while ( myMotorDriver.ready() == false );


  //*****Set application settings and enable driver*****//

  while ( myMotorDriver.busy() );
  myMotorDriver.enable();

#ifdef BLE_SHOW_DATA
  //SERIAL_PORT.begin(115200);
  //delay(1000);
  //SERIAL_PORT.printf("Viper. Compiled: %s\n" , __TIME__);
#endif

#ifdef AM_DEBUG_PRINTF
  enable_print_interface();
#endif
  analogWriteResolution(16); //Set AnalogWrite resolution to 16 bit = 0 - 65535 (but make max 64k or trouble)
  set_Adv_Name(); //BLE_example_funcs.cpp near end of file, fn declared extern in BLE_example.h
  HciDrvRadioBoot(0);
  exactle_stack_init();
  HciVsA3_SetRfPowerLevelEx(TX_POWER_LEVEL_PLUS_3P0_dBm); //= 15 decimal = max power WORKS..default = 0


  AmdtpStart();

  pinMode(LED_BUILTIN, OUTPUT);

  //Set a starting point...for motors, servo, and LED_BUILTIN
  //delay(1000);

  // Bluetooth would start after blinking
  for (int i = 0; i < 20; i++)
  {
    digitalWrite(LED_BUILTIN, HIGH);
    delay(50);
    digitalWrite(LED_BUILTIN, LOW);
    delay(50);
  }

  //setupwdt();

  pinMode(blinkPin, OUTPUT);
  digitalWrite(blinkPin, LOW);

  // Configure the watchdog.
  //setupTimerA(myTimer, 31); // timerNum, period - //moved to BLE_example_funcs.cpp scheduler_timer_init
  setupWdt();
  am_hal_wdt_init(&g_sWatchdogConfig);
  //NVIC_EnableIRQ(CTIMER_IRQn); // Enable CTIMER interrupt in nested vector interrupt controller.
  NVIC_EnableIRQ(WDT_IRQn); // Enable WDT interrupt in nested vector interrupt controller.

  uint8_t a = 0;
  m_Rcvd = &a;
  ////Serial.printf("Size of command: %d", sizeof(cmd_t));
  am_hal_interrupt_master_enable();
  //interrupts(); // Enable interrupt operation. Equivalent to am_hal_rtc_int_enable().
  //am_hal_wdt_start();
  //am_hal_wdt_int_enable(); - freezes boot
  while (!SERIAL_PORT) {};

#ifdef USE_SPI
  SPI_PORT.begin();
#else
  WIRE_PORT.begin();
  WIRE_PORT.setClock(400000);
#endif

  bool initialized = false;
  while ( !initialized ) {

#ifdef USE_SPI
    myICM.begin( CS_PIN, SPI_PORT );
#else
    myICM.begin( WIRE_PORT, AD0_VAL );
#endif

    SERIAL_PORT.print( F("Initialization of the sensor returned: ") );
    SERIAL_PORT.println( myICM.statusString() );
    if ( myICM.status != ICM_20948_Stat_Ok ) {
      SERIAL_PORT.println( "Trying again..." );
      delay(500);
    } else {
      initialized = true;
    }
  }

  myPID.SetMode(AUTOMATIC);



} /*** END setup FCN ***/

int counter = 0;
float i = 0, j = 0;
float yaw = 0, vel;
float last_time;
float dt, motor_speed;

void loop()
{
  if (l_Rcvd > 1) //Check if we have a new message from amdtps_main.c through BLE_example_funcs.cpp
  {
    cmd = (cmd_t *)m_Rcvd;
    switch (cmd->command_type)
    {
      case SET_MOTORS:
        //Serial.println("Placeholder: Set Motors");
        break;
      case GET_MOTORS:
        //Serial.println("Placeholder: Set Motors");
        //amdtpsSendData((uint8_t *)res_cmd, *val_len);
        break;
      case SER_RX:
        //Serial.println("Got a serial message");
        pushMessage((char *)&cmd->data, cmd->length);
        break;
      case REQ_FLOAT:
        //Serial.println("Going to send a float");
        //TODO: Put a float (perhaps pi) into a command response and send it.
        res_cmd->command_type = GIVE_FLOAT;
        res_cmd-> length = 4;
        ((float *)(res_cmd->data))[0] = 5.51f;
        amdtpsSendData((uint8_t *)res_cmd, 6);
        break;
      case PING:
        //Serial.println("Ping Pong");
        cmd->command_type = PONG;
        amdtpsSendData(m_Rcvd, l_Rcvd);
        break;
      case START_BYTESTREAM_TX:
        bytestream_active = 1; //(int)cmd->data[0];
        ////Serial.printf("Start bytestream with active %d \n", bytestream_active);
        ((uint32_t *)res_cmd->data)[0] = 0;
        break;
      case STOP_BYTESTREAM_TX:
        bytestream_active = 0;
        break;
      default:
        //Serial.printf("Unsupported Command 0x%x \n", cmd->command_type);
        break;
    }

    l_Rcvd = 0;
    am_hal_wdt_restart();
    free(m_Rcvd);
  } //End if s_Rcvd != 100
  else if ((s_Rcvd[0] == '6' && s_Rcvd[1] == '7'))
  {
    s_Rcvd[0] = 0;
    digitalWrite(LED_BUILTIN, HIGH);
    ////Serial.printf("Connected, length was %d", l_Rcvd);
  }
  else if ((s_Rcvd[0] == '6' && s_Rcvd[1] == '8'))
  {
    digitalWrite(LED_BUILTIN, LOW);
    //Serial.println("disconnected");
    //Decimal value of D for Disconnect
    ////Serial.println("got disconnect from case in ino file - set_Stop");
    digitalWrite(LED_BUILTIN, LOW);
    //amdtps_conn_close();
    DmDevReset();
  }

  if (bytestream_active)
  {
    //sending data
    res_cmd->command_type = BYTESTREAM_TX;
    res_cmd-> length = 70;
    memcpy(res_cmd->data, &last_time, 4);
    memcpy(res_cmd->data + 4, &motor_speed, 4);
    memcpy(res_cmd->data + 8, &vel,4);
    amdtpsSendData((uint8_t *)res_cmd, 14);
    
    //taking measurements & perform PID
    if (myICM.dataReady()) {
      myICM.getAGMT();
      vel = myICM.gyrZ();
      Input = vel;
      Setpoint = 20;
      myPID.Compute();

      //update the motor power
      motor_speed = min(60, Output + 50);
      
      myMotorDriver.setDrive(0, 0, motor_speed);
      myMotorDriver.setDrive(1, 0, motor_speed);

      //getting time
      last_time = micros();

    }

  }
  // ---------------------------------------------------------------------------
  trigger_timers();
  delay(10);
} //END LOOP
